@@ -1,7 +1,7 @@
 #!/usr/bin/env python2
 
 # pyinotify.py - python interface to inotify
-# Copyright (c) 2005-2011 Sebastien Martini <seb@dbzteam.org>
+# Copyright (c) 2005-2015 Sebastien Martini <seb@dbzteam.org>
 #
 # Permission is hereby granted, free of charge, to any person obtaining a copy
 # of this software and associated documentation files (the "Software"), to deal
@@ -68,7 +68,6 @@ from datetime import datetime, timedelta
 import time
 import re
 import asyncore
-import glob
 import subprocess
 
 try:
@@ -76,6 +75,12 @@ try:
 except ImportError:
     pass  # Will fail on Python 2.4 which has reduce() builtin anyway.
 
+try:
+    from glob import iglob as glob
+except ImportError:
+    # Python 2.4 does not have glob.iglob().
+    from glob import glob as glob
+
 try:
     import ctypes
     import ctypes.util
@@ -90,7 +95,7 @@ except ImportError:
 
 __author__ = "seb@dbzteam.org (Sebastien Martini)"
 
-__version__ = "0.9.4"
+__version__ = "0.9.6"
 
 __metaclass__ = type  # Use new-style classes by default
 
@@ -197,9 +202,14 @@ class _CtypesLibcINotifyWrapper(INotifyWrapper):
 
     def init(self):
         assert ctypes
+
+        try_libc_name = 'c'
+        if sys.platform.startswith('freebsd'):
+            try_libc_name = 'inotify'
+
         libc_name = None
         try:
-            libc_name = ctypes.util.find_library('c')
+            libc_name = ctypes.util.find_library(try_libc_name)
         except (OSError, IOError):
             pass  # Will attemp to load it with None anyway.
 
@@ -223,7 +233,7 @@ class _CtypesLibcINotifyWrapper(INotifyWrapper):
 
         self._libc.inotify_init.argtypes = []
         self._libc.inotify_init.restype = ctypes.c_int
-        self._libc.inotify_add_watch.argtypes = [ctypes.c_int, ctypes.c_char_p, 
+        self._libc.inotify_add_watch.argtypes = [ctypes.c_int, ctypes.c_char_p,
                                                  ctypes.c_uint32]
         self._libc.inotify_add_watch.restype = ctypes.c_int
         self._libc.inotify_rm_watch.argtypes = [ctypes.c_int, ctypes.c_int]
@@ -248,10 +258,6 @@ class _CtypesLibcINotifyWrapper(INotifyWrapper):
         assert self._libc is not None
         return self._libc.inotify_rm_watch(fd, wd)
 
-    def _sysctl(self, *args):
-        assert self._libc is not None
-        return self._libc.sysctl(*args)
-
 
 # Logging
 def logger_init():
@@ -268,40 +274,18 @@ log = logger_init()
 
 
 # inotify's variables
-class SysCtlINotify:
+class ProcINotify:
     """
-    Access (read, write) inotify's variables through sysctl. Usually it
-    requires administrator rights to update them.
+    Access (read, write) inotify's variables through /proc/sys/. Note that
+    usually it requires administrator rights to update them.
 
     Examples:
       - Read max_queued_events attribute: myvar = max_queued_events.value
       - Update max_queued_events attribute: max_queued_events.value = 42
     """
-
-    inotify_attrs = {'max_user_instances': 1,
-                     'max_user_watches': 2,
-                     'max_queued_events': 3}
-
-    def __init__(self, attrname, inotify_wrapper):
-        # FIXME: right now only supporting ctypes
-        assert ctypes
-        self._attrname = attrname
-        self._inotify_wrapper = inotify_wrapper
-        sino = ctypes.c_int * 3
-        self._attr = sino(5, 20, SysCtlINotify.inotify_attrs[attrname])
-
-    @staticmethod
-    def create(attrname):
-        """
-        Factory method instanciating and returning the right wrapper.
-        """
-        # FIXME: right now only supporting ctypes
-        if ctypes is None:
-            return None
-        inotify_wrapper = _CtypesLibcINotifyWrapper()
-        if not inotify_wrapper.init():
-            return None
-        return SysCtlINotify(attrname, inotify_wrapper)
+    def __init__(self, attr):
+        self._base = "/proc/sys/fs/inotify"
+        self._attr = attr
 
     def get_val(self):
         """
@@ -309,14 +293,14 @@ class SysCtlINotify:
 
         @return: stored value.
         @rtype: int
+        @raise IOError: if corresponding file in /proc/sys cannot be read.
         """
-        oldv = ctypes.c_int(0)
-        size = ctypes.c_int(ctypes.sizeof(oldv))
-        self._inotify_wrapper._sysctl(self._attr, 3,
-                                      ctypes.c_voidp(ctypes.addressof(oldv)),
-                                      ctypes.addressof(size),
-                                      None, 0)
-        return oldv.value
+        file_obj = file(os.path.join(self._base, self._attr), 'r')
+        try:
+            val = int(file_obj.readline())
+        finally:
+            file_obj.close()
+        return val
 
     def set_val(self, nval):
         """
@@ -324,33 +308,31 @@ class SysCtlINotify:
 
         @param nval: replaces current value by nval.
         @type nval: int
+        @raise IOError: if corresponding file in /proc/sys cannot be written.
         """
-        oldv = ctypes.c_int(0)
-        sizeo = ctypes.c_int(ctypes.sizeof(oldv))
-        newv = ctypes.c_int(nval)
-        sizen = ctypes.c_int(ctypes.sizeof(newv))
-        self._inotify_wrapper._sysctl(self._attr, 3,
-                                      ctypes.c_voidp(ctypes.addressof(oldv)),
-                                      ctypes.addressof(sizeo),
-                                      ctypes.c_voidp(ctypes.addressof(newv)),
-                                      ctypes.addressof(sizen))
+        file_obj = file(os.path.join(self._base, self._attr), 'w')
+        try:
+            file_obj.write(str(nval) + '\n')
+        finally:
+            file_obj.close()
 
     value = property(get_val, set_val)
 
     def __repr__(self):
-        return '<%s=%d>' % (self._attrname, self.get_val())
+        return '<%s=%d>' % (self._attr, self.get_val())
 
 
 # Inotify's variables
 #
-# FIXME: currently these variables are only accessible when ctypes is used,
-#        otherwise there are set to None.
+# Note: may raise IOError if the corresponding value in /proc/sys
+#       cannot be accessed.
 #
-# read: myvar = max_queued_events.value
-# update: max_queued_events.value = 42
+# Examples:
+#  - read: myvar = max_queued_events.value
+#  - update: max_queued_events.value = 42
 #
 for attrname in ('max_queued_events', 'max_user_instances', 'max_user_watches'):
-    globals()[attrname] = SysCtlINotify.create(attrname)
+    globals()[attrname] = ProcINotify(attrname)
 
 
 class EventsCodes:
@@ -730,23 +712,28 @@ class _SysProcessEvent(_ProcessEvent):
                 # Since the directory d2 is new, then everything inside it must
                 # also be new.
                 created_dir_wd = addw_ret.get(created_dir)
-                if (created_dir_wd is not None) and (created_dir_wd > 0):
-                    for name in os.listdir(created_dir):
-                        inner = os.path.join(created_dir, name)
-                        if self._watch_manager.get_wd(inner) is not None:
-                            continue
-                        # Generate (simulate) creation events for sub-
-                        # directories and files.
-                        if os.path.isfile(inner):
-                            # symlinks are handled as files.
-                            flags = IN_CREATE
-                        elif os.path.isdir(inner):
-                            flags = IN_CREATE | IN_ISDIR
-                        else:
-                            # This path should not be taken.
-                            continue
-                        rawevent = _RawEvent(created_dir_wd, flags, 0, name)
-                        self._notifier.append_event(rawevent)
+                if ((created_dir_wd is not None) and (created_dir_wd > 0) and
+                    os.path.isdir(created_dir)):
+                    try:
+                        for name in os.listdir(created_dir):
+                            inner = os.path.join(created_dir, name)
+                            if self._watch_manager.get_wd(inner) is not None:
+                                continue
+                            # Generate (simulate) creation events for sub-
+                            # directories and files.
+                            if os.path.isfile(inner):
+                                # symlinks are handled as files.
+                                flags = IN_CREATE
+                            elif os.path.isdir(inner):
+                                flags = IN_CREATE | IN_ISDIR
+                            else:
+                                # This path should not be taken.
+                                continue
+                            rawevent = _RawEvent(created_dir_wd, flags, 0, name)
+                            self._notifier.append_event(rawevent)
+                    except OSError, err:
+                        msg = "process_IN_CREATE, invalid directory %s: %s"
+                        log.debug(msg % (created_dir, str(err)))
         return self.process_default(raw_event)
 
     def process_IN_MOVED_FROM(self, raw_event):
@@ -1126,8 +1113,8 @@ class Notifier:
         @type default_proc_fun: instance of ProcessEvent
         @param read_freq: if read_freq == 0, events are read asap,
                           if read_freq is > 0, this thread sleeps
-                          max(0, read_freq - timeout) seconds. But if
-                          timeout is None it may be different because
+                          max(0, read_freq - (timeout / 1000)) seconds. But
+                          if timeout is None it may be different because
                           poll is blocking waiting for something to read.
         @type read_freq: int
         @param threshold: File descriptor will be read only if the accumulated
@@ -1138,8 +1125,9 @@ class Notifier:
                           until the amount of events to read is >= threshold.
                           At least with read_freq set you might sleep.
         @type threshold: int
-        @param timeout:
-            http://docs.python.org/lib/poll-objects.html#poll-objects
+        @param timeout: see read_freq above. If provided, it must be set in
+                        milliseconds. See
+                        https://docs.python.org/2/library/select.html#polling-objects
         @type timeout: int
         """
         # Watch Manager instance
@@ -1205,7 +1193,8 @@ class Notifier:
         milliseconds.
 
         @param timeout: If specified it overrides the corresponding instance
-                        attribute _timeout.
+                        attribute _timeout. timeout must be sepcified in
+                        milliseconds.
         @type timeout: int
 
         @return: New events to read.
@@ -1279,6 +1268,9 @@ class Notifier:
         """
         while self._eventq:
             raw_event = self._eventq.popleft()  # pop next event
+            if self._watch_manager.ignore_events:
+                log.debug("Event ignored: %s" % repr(raw_event))
+                continue
             watch_ = self._watch_manager.get_watch(raw_event.wd)
             if (watch_ is None) and not (raw_event.mask & IN_Q_OVERFLOW):
                 if not (raw_event.mask & IN_IGNORED):
@@ -1355,7 +1347,6 @@ class Notifier:
             # Register unlink function
             atexit.register(lambda : os.unlink(pid_file))
 
-
     def _sleep(self, ref_time):
         # Only consider sleeping if read_freq is > 0
         if self._read_freq > 0:
@@ -1365,7 +1356,6 @@ class Notifier:
                 log.debug('Now sleeping %d seconds', sleep_amount)
                 time.sleep(sleep_amount)
 
-
     def loop(self, callback=None, daemonize=False, **args):
         """
         Events are read only one time every min(read_freq, timeout)
@@ -1412,15 +1402,18 @@ class Notifier:
         # Close internals
         self.stop()
 
-
     def stop(self):
         """
         Close inotify's instance (close its file descriptor).
         It destroys all existing watches, pending events,...
         This method is automatically called at the end of loop().
+        Afterward it is invalid to access this instance.
         """
-        self._pollobj.unregister(self._fd)
-        os.close(self._fd)
+        if self._fd is not None:
+            self._pollobj.unregister(self._fd)
+            os.close(self._fd)
+            self._fd = None
+        self._sys_proc_fun = None
 
 
 class ThreadedNotifier(threading.Thread, Notifier):
@@ -1444,7 +1437,7 @@ class ThreadedNotifier(threading.Thread, Notifier):
         @type default_proc_fun: instance of ProcessEvent
         @param read_freq: if read_freq == 0, events are read asap,
                           if read_freq is > 0, this thread sleeps
-                          max(0, read_freq - timeout) seconds.
+                          max(0, read_freq - (timeout / 1000)) seconds.
         @type read_freq: int
         @param threshold: File descriptor will be read only if the accumulated
                           size to read becomes >= threshold. If != 0, you likely
@@ -1454,8 +1447,9 @@ class ThreadedNotifier(threading.Thread, Notifier):
                           until the amount of events to read is >= threshold. At
                           least with read_freq you might sleep.
         @type threshold: int
-        @param timeout:
-           see http://docs.python.org/lib/poll-objects.html#poll-objects
+        @param timeout: see read_freq above. If provided, it must be set in
+                        milliseconds. See
+                        https://docs.python.org/2/library/select.html#select.poll.poll
         @type timeout: int
         """
         # Init threading base class
@@ -1545,28 +1539,32 @@ class TornadoAsyncNotifier(Notifier):
     Tornado ioloop adapter.
 
     """
-    def __init__(self, watch_manager, ioloop, callback=None, 
-                 default_proc_fun=None, read_freq=0, threshold=0, timeout=None, 
+    def __init__(self, watch_manager, ioloop, callback=None,
+                 default_proc_fun=None, read_freq=0, threshold=0, timeout=None,
                  channel_map=None):
         """
-        Note that if later you must call ioloop.close() be sure to let the 
+        Note that if later you must call ioloop.close() be sure to let the
         default parameter to all_fds=False.
 
         See example tornado_notifier.py for an example using this notifier.
 
         @param ioloop: Tornado's IO loop.
         @type ioloop: tornado.ioloop.IOLoop instance.
-        @param callback: Functor called at the end of each call to handle_read 
-                         (IOLoop's read handler). Expects to receive the 
+        @param callback: Functor called at the end of each call to handle_read
+                         (IOLoop's read handler). Expects to receive the
                          notifier object (self) as single parameter.
         @type callback: callable object or function
         """
-        self.io_loop = ioloop 
+        self.io_loop = ioloop
         self.handle_read_callback = callback
         Notifier.__init__(self, watch_manager, default_proc_fun, read_freq,
                           threshold, timeout)
         ioloop.add_handler(self._fd, self.handle_read, ioloop.READ)
 
+    def stop(self):
+        self.io_loop.remove_handler(self._fd)
+        Notifier.stop(self)
+
     def handle_read(self, *args, **kwargs):
         """
         See comment in AsyncNotifier.
@@ -1578,6 +1576,43 @@ class TornadoAsyncNotifier(Notifier):
             self.handle_read_callback(self)
 
 
+class AsyncioNotifier(Notifier):
+    """
+
+    asyncio/trollius event loop adapter.
+
+    """
+    def __init__(self, watch_manager, loop, callback=None,
+                 default_proc_fun=None, read_freq=0, threshold=0, timeout=None):
+        """
+
+        See examples/asyncio_notifier.py for an example usage.
+
+        @param loop: asyncio or trollius event loop instance.
+        @type loop: asyncio.BaseEventLoop or trollius.BaseEventLoop instance.
+        @param callback: Functor called at the end of each call to handle_read.
+                         Expects to receive the notifier object (self) as
+                         single parameter.
+        @type callback: callable object or function
+
+        """
+        self.loop = loop
+        self.handle_read_callback = callback
+        Notifier.__init__(self, watch_manager, default_proc_fun, read_freq,
+                          threshold, timeout)
+        loop.add_reader(self._fd, self.handle_read)
+
+    def stop(self):
+        self.loop.remove_reader(self._fd)
+        Notifier.stop(self)
+
+    def handle_read(self, *args, **kwargs):
+        self.read_events()
+        self.process_events()
+        if self.handle_read_callback is not None:
+            self.handle_read_callback(self)
+
+
 class Watch:
     """
     Represent a watch, i.e. a file or directory being watched.
@@ -1639,11 +1674,11 @@ class ExcludeFilter:
     def __init__(self, arg_lst):
         """
         Examples:
-          ef1 = ExcludeFilter(["^/etc/rc.*", "^/etc/hostname"])
+          ef1 = ExcludeFilter(["/etc/rc.*", "/etc/hostname"])
           ef2 = ExcludeFilter("/my/path/exclude.lst")
           Where exclude.lst contains:
-          ^/etc/rc.*
-          ^/etc/hostname
+          /etc/rc.*
+          /etc/hostname
 
         Note: it is not possible to exclude a file if its encapsulating
               directory is itself watched. See this issue for more details
@@ -1734,6 +1769,7 @@ class WatchManager:
                                filter for every call to add_watch.
         @type exclude_filter: callable object
         """
+        self._ignore_events = False
         self._exclude_filter = exclude_filter
         self._wmd = {}  # watch dict key: watch descriptor, value: watch
 
@@ -1830,7 +1866,7 @@ class WatchManager:
 
     def __glob(self, path, do_glob):
         if do_glob:
-            return glob.iglob(path)
+            return glob(path)
         else:
             return [path]
 
@@ -1845,7 +1881,7 @@ class WatchManager:
         although unicode paths are accepted there are converted to byte
         strings before a watch is put on that path. The encoding used for
         converting the unicode object is given by sys.getfilesystemencoding().
-        If |path| si already watched it is ignored, but if it is called with
+        If |path| is already watched it is ignored, but if it is called with
         option rec=True a watch is put on each one of its not-watched
         subdirectory.
 
@@ -2166,6 +2202,16 @@ class WatchManager:
                               auto_add=False, do_glob=False,
                               exclude_filter=lambda path: False)
 
+    def get_ignore_events(self):
+        return self._ignore_events
+
+    def set_ignore_events(self, nval):
+        self._ignore_events = nval
+
+    ignore_events = property(get_ignore_events, set_ignore_events,
+                             "Make watch manager ignoring new events.")
+
+
 
 class RawOutputFormat:
     """
@@ -2253,6 +2299,9 @@ def command_line():
     parser.add_option("-a", "--auto_add", action="store_true",
                       dest="auto_add",
                       help="Automatically add watches on new directories")
+    parser.add_option("-g", "--glob", action="store_true",
+                      dest="glob",
+                      help="Treat paths as globs")
     parser.add_option("-e", "--events-list", metavar="EVENT[,...]",
                       dest="events_list",
                       help=("A comma-separated list of events to watch for - "
@@ -2328,7 +2377,7 @@ def command_line():
 
     log.debug('Start monitoring %s, (press c^c to halt pyinotify)' % path)
 
-    wm.add_watch(path, mask, rec=options.recursive, auto_add=options.auto_add)
+    wm.add_watch(path, mask, rec=options.recursive, auto_add=options.auto_add, do_glob=options.glob)
     # Loop forever (until sigint signal get caught)
     notifier.loop(callback=cb_fun)
 
